APPLICATION_X_AVRO_BINARY = "application/x-avro-binary"
APPLICATION_X_AVRO_JSON = "application/x-avro-json"
APPLICATION_X_JSON_SCHEMA = "application/x-json-schema"
APPLICATION_JSON = "application/json"

AVRO_SERIALIZATION_TYPES = (
    APPLICATION_X_AVRO_BINARY,
    APPLICATION_X_AVRO_JSON,
)
SERIALIZATION_TYPES = AVRO_SERIALIZATION_TYPES + (
    APPLICATION_X_JSON_SCHEMA,
    APPLICATION_JSON,
)
